package net;


import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.SocketException;
import java.nio.ByteBuffer;

public class AppSocket extends Socket{
	private ArrayBlockingQueue<ByteBuffer> bufferToApp;
	private ArrayBlockingQueue<Integer> ackBuffer;

	private final int BUFFER_SIZE = 5;
	private final int TIMEOUT_TIME = 800;
	private final int TIMEOUT_LIMIT = 3;
	private byte seqNum;
	private byte expectedSeq;
	private boolean vive;
	public AppSocket(Transport transport, int id){
		super(transport, id);
		bufferToApp = new  ArrayBlockingQueue<ByteBuffer>(BUFFER_SIZE);
		ackBuffer = new  ArrayBlockingQueue<Integer>(BUFFER_SIZE);
		this.seqNum = 0;
		this.expectedSeq = 0;
		vive = true;
	}
	public void kill(){
		byte[] dados = {-33};
		this.vive = false;
		DatagramPacket pacoteKill = new DatagramPacket(dados, dados.length);
		try {
			super.buffer.put(pacoteKill);
		} catch (InterruptedException e) {
			System.out.println(e.getMessage());
		}
	}
	public void connect(ConnectionInfo server) throws TimeoutLimitException{
		this.seqNum = 0;
		this.expectedSeq = 0;
		super.transport.addListenerSocket(this);
		transport.sendRequest(server, super.id);
		boolean recebeu = false;
		int contador = 0;
		DatagramPacket pacote = null;
		while(!recebeu){
			pacote = this.getFromMainBuffer();
			if(pacote != null){
				recebeu  = true; 
			}else{
				contador++;
				if(contador >= TIMEOUT_LIMIT){
					throw new TimeoutLimitException();
				}
				transport.sendRequest(server, super.id);
			}
		}
		InetAddress ip = pacote.getAddress();
		byte[] temp = pacote.getData();
		int id = ByteOp.byteArrayToInt(new byte[]{temp[4], temp[5], temp[6], temp[7]});
		int porta = ByteOp.byteArrayToInt(new byte[]{temp[8],temp[9], temp[10], temp[11]});
		ConnectionInfo ci = new ConnectionInfo(id, ip, porta);
		transport.addSocket(this, ci);
		this.start();
	}
	public void match(ConnectionInfo info){
		this.start();
		transport.addSocket(this, info);
	}
	@Override
	public void run(){
		while(vive){
			byte[] temp = null;
			try {
				temp = super.buffer.take().getData();
			} catch (InterruptedException e) {
				System.out.println(e.getMessage());
			}
			
			if(temp != null){
				if (temp[0] != -33) {
					if (temp[12] == -1) {
						byte seqRecebido = temp[13];
						if (expectedSeq != seqRecebido) {
							System.out.println("AppSocket "+this.id+" Recebeu pacote duplicado");
							this.sendAck(seqRecebido);
						} else {
							bufferToApp.offer(ByteBuffer.wrap(temp));
							this.sendAck(expectedSeq);
							expectedSeq = (byte) (expectedSeq ^ 1);
						}
					} else {
						ackBuffer.offer(new Integer(temp[12]));
					}
				}else{
					bufferToApp.offer(ByteBuffer.wrap(temp)); // Manda o kill pra quem estiver esperando
				}
			}
		}
		transport.removeSocket(this);
	}

	public void send(byte[] dados) throws TimeoutLimitException {
		byte[] pacote = this.makePkt(dados);
		int contador = 0;
		boolean mandou = false;
		this.transport.send(pacote, super.id);
		while(!mandou){
			
			Integer temp = this.getFromAckBuffer();
			if(temp != null){
				mandou = true; 
			}else{
				contador++;
				if(contador >= TIMEOUT_LIMIT){
					throw new TimeoutLimitException();
				}
				this.transport.send(pacote, super.id);
			}
		}
	}
	public void sendProbe(ConnectionInfo ci) throws TimeoutLimitException{
		int contador = 0;
		boolean mandou = false;
		this.transport.sendProbe(ci, super.id);
		while(!mandou){
			
			Integer temp = this.getFromAckBuffer();
			if(temp != null){
				mandou = true; 
			}else{
				contador++;
				if(contador >= TIMEOUT_LIMIT){
					throw new TimeoutLimitException();
				}
				this.transport.sendProbe(ci, super.id);
			}
		}
	}
		
	public byte[] receive(int timeOut){
		ByteBuffer temp = null;
		byte[] retorno = null;
		try {
			temp = bufferToApp.poll(timeOut, TimeUnit.MILLISECONDS);
		} catch (InterruptedException e) {
			System.out.println(e.getMessage());
		}
		if(temp != null){
			if(temp.get(0) != -33){
				temp.position(14);
				int tamanho = ByteOp.byteArrayToInt(new byte[]{temp.get(), temp.get(), temp.get(), temp.get()});
				retorno = new byte[tamanho];
				for(int i = 0; i < tamanho; i++){
					retorno[i] = temp.get();
				}
			}else{
				retorno = new byte[]{-33};
			}
		}
		
		return retorno;
	}
	public byte[] receive(){
		ByteBuffer temp = null;
		byte[] retorno = null;
		try {
			temp = bufferToApp.take();
		} catch (InterruptedException e) {
			System.out.println(e.getMessage());
		}
		if(temp != null){
			if(temp.get(0) != -33){
				temp.position(14);
				int tamanho = ByteOp.byteArrayToInt(new byte[]{temp.get(), temp.get(), temp.get(), temp.get()});
				retorno = new byte[tamanho];
				for(int i = 0; i < tamanho; i++){
					retorno[i] = temp.get();
				}
			}else{
				retorno = new byte[]{-33};
			}
		}
		
		return retorno;
	}
	

	/*
	 * Dados de no max 1006 bytes
	 * + 4 bytes do tamanho
	 * + 1 byte ack
	 * + 1 byte seq
	 * total = 1012 bytes max
	 */
	public byte[] makePkt(byte[] data){
		ByteBuffer bf = ByteBuffer.allocate(data.length+6);
		byte[] dataSize = ByteOp.intToByte(data.length);
		byte ack = -1;//pacote de dados
		bf.put(ack);
		bf.put(seqNum);
		bf.put(dataSize);
		bf.put(data);
		bf.flip();
		seqNum = (byte)(seqNum ^1);
		return bf.array();
		
	}
	public void acceptConnection(){
		byte[] pacote = this.makePkt(new byte[]{});
		seqNum = (byte)(seqNum ^1); // volta o numero de sequencia
		this.transport.send(pacote, this.id);
	}
	private void sendAck(byte ackRecebido){
		byte[] ack = {ackRecebido};
		this.transport.send(ack, super.id);
	}
	private DatagramPacket getFromMainBuffer(){
		DatagramPacket ret = null;
		try{
			ret = super.buffer.poll(TIMEOUT_TIME, TimeUnit.MILLISECONDS);
			
		}catch (InterruptedException e) {
			System.out.println("Fui interrompido: "+e.getMessage());
		}
		return ret;
	}
	private Integer getFromAckBuffer(){
		Integer ret = null;
		try {
			ret = ackBuffer.poll(TIMEOUT_TIME, TimeUnit.MILLISECONDS);
		} catch (InterruptedException e) {
			System.out.println("Fui interrompido: "+e.getMessage());
		}
		return ret;
	}

	public static void main(String[] args) throws SocketException, InterruptedException {
		Transport trans1 = new Transport(5000);
		AppSocket testaMorte = new AppSocket(trans1, 1);
		testaMorte.start();
		Thread.sleep(1000);
		testaMorte.kill();
		trans1.kill();
		
	}
}
